from typing import Iterable

import jax.numpy as jnp

from moll.core.distance import (
    add_point_to_bag,
    add_points_to_bag,
    is_in_bag,
    keep_last_changes,
)

__all__ = ["OnlineDiversityPicker"]


class OnlineDiversityPicker:
    def __init__(
        self,
        n_points: int,
        dist: callable,
        n_neighbors: int = 4,
        threshold: float = 0.0,
    ):
        self.n_points: int = n_points
        self.dist = dist
        self.n_neighbors = n_neighbors
        self.threshold = threshold

        self.data_ = None
        self.labels = []

        self.n_seen: int = 0
        self.n_accepted: int = 0

    def _fill_if_not_full(self, point: jnp.ndarray) -> (bool, int):
        """
        Add point if pickier is not filed yet.
        """

        idx = None

        # If it's a first point, just add it
        if self.is_empty():
            self.data_ = jnp.array([point])
            return True, 0

        # If it's not a first point, check if it's already in the bag
        if is_accepted := not is_in_bag(point, self.data_):
            self.data_ = jnp.vstack([self.data_, point])
            idx = len(self.data_) - 1

        return is_accepted, idx

    def _create_label(self) -> int:
        """
        Return the next label.
        """
        return self.n_seen

    def _add_label_for_accepted_point(self, label, idx) -> None:
        """
        Add a label to the picker.
        """

        label = label or self._create_label()

        assert label not in self.labels, f"Label `{label}` is already in the picker"

        if (idx is None) or (idx == len(self.labels)):
            self.labels.append(label)
        else:
            self.labels[idx] = label

    def append(self, point, label=None, return_idx=False) -> bool:
        """
        Add a point to the picker.
        """

        # If we haven't seen enough points, just add the point to the data rejecting duplicates
        if not self.is_full():
            is_accepted, old_idx = self._fill_if_not_full(point)
        else:
            # If we have seen enough points, decide whether to add the point or not
            self.data_, is_accepted, old_idx = add_point_to_bag(
                x=point,
                X=self.data_,
                dist=self.dist,
                n_neighbors=self.n_neighbors,
                threshold=self.threshold,
            )

            old_idx = None if old_idx < 0 else old_idx

        # Update the number of viewed points
        self._update_counters(is_accepted, label, old_idx)

        # Return whether the point was accepted or not
        if return_idx:
            return is_accepted, old_idx
        return is_accepted

    def fast_init(self, points: jnp.ndarray) -> None:
        """
        Initialize the picker with a batch of points.
        """

        assert len(points) == self.n_points

        self.data_ = jnp.array(points)
        self.n_accepted = len(points)
        self.n_seen = len(points)

    def extend(self, points: jnp.ndarray | list, labels=None, *, pairs=None) -> int:
        """
        Add a batch of points to the picker.
        """

        assert pairs is None or (
            points is None and labels is None
        ), "`pairs` are provided, then `points` and `labels` must be None"

        if pairs:
            points, labels = zip(*pairs)

        changed_item_idxs = -jnp.ones(len(points), dtype=int)

        # STAGE 1: If we haven't seen enough points, just add the points to the data rejecting duplicates

        for idx, point in enumerate(points):
            if self.is_full():
                break
            else:
                label = labels and labels[idx]
                is_accepted, old_idx = self.append(point, label=label, return_idx=True)
                if is_accepted:
                    changed_item_idxs = changed_item_idxs.at[idx].set(old_idx)
        else:
            # If we iterated over all points, return the number of added points
            return (changed_item_idxs >= 0).sum().item()

        # STAGE 2: If there still some point:
        points = points[idx:]

        self.data_, accepted_points_mask2, changed_item_idxs2 = add_points_to_bag(
            xs=points,
            X=self.data_,
            dist=self.dist,
            n_neighbors=self.n_neighbors,
            threshold=self.threshold,
        )

        # Update the number of accepted points
        for new_point_idx, (is_accepted, replaced_point_idx) in enumerate(
            zip(accepted_points_mask2, changed_item_idxs2), start=idx
        ):
            label = labels and labels[new_point_idx]
            self._update_counters(is_accepted, label, replaced_point_idx)

        # STAGE 3: Combine the results of the two stages

        changed_item_idxs = changed_item_idxs.at[idx:].set(changed_item_idxs2)

        # If during the first stage some points were added
        if (changed_item_idxs[:idx] >= 0).any():
            # Some points might have been accepted when the picker was not full and then
            # replaced by another point. In that case we counted them twice, we need to
            # decrement the number of accepted points

            # Now `changed_item_idxs` contains changes from both stages
            n_accepted_on_both_stages = (changed_item_idxs >= 0).sum()

            # Keep only the last unique changes
            changed_item_idxs = keep_last_changes(changed_item_idxs)
            n_unique_changes = (changed_item_idxs >= 0).sum()

            # We can't accept more points than the number of points we need to pick
            n_could_be_accepted = min(self.n_points, n_unique_changes)

            # Calculate the number of extra accepted points
            n_extra_accepted = n_accepted_on_both_stages - n_could_be_accepted

            # Decrease the number of accepted points
            self.n_accepted -= n_extra_accepted.item()

        accepted_points_mask = changed_item_idxs >= 0
        n_accepted = accepted_points_mask.sum().item()

        return n_accepted

    def _update_counters(self, is_accepted: bool, label=None, label_idx=None):
        """
        Update the counters.
        """
        if is_accepted:
            self.n_accepted += 1
            self._add_label_for_accepted_point(label, label_idx)
        self.n_seen += 1

    @property
    def n_rejected(self) -> int:
        """
        Return the number of rejected points.
        """
        return self.n_seen - self.n_accepted

    def size(self) -> int:
        """
        Return the number of points in the picker.
        """
        return self.data_.shape[0]

    def is_full(self) -> bool:
        """
        Return whether the picker is full or not.
        """
        if self.is_empty():
            return False
        return self.size() >= self.n_points

    def is_empty(self) -> bool:
        """
        Return whether the picker is empty or not.
        """
        return self.data_ is None

    def labels(self) -> list:
        """
        Return the currently picked labels.
        """
        return self.labels

    @property
    def points(self) -> list:
        """
        Return the currently picked points.
        """
        return self.data_

    @property
    def dim(self) -> int:
        """
        Return the dimension of the points.
        """
        if self.is_empty():
            return None
        return self.data_.shape[1]



import itertools

import jax
import jax.numpy as jnp


def points_around(
    center: jnp.ndarray,
    n_points: int,
    key: jnp.ndarray,
    radius: float | int = 1,
):
    dim = len(center)
    offsets = jax.random.uniform(
        key, shape=(n_points, dim), minval=-radius, maxval=radius
    )
    return center + offsets


def generate_points(
    centers: jnp.ndarray,
    sizes: list,
    seed: int,
    radius: float | int = 1,
    shuffle: bool = True,
) -> jnp.ndarray:
    """
    Generate points around centers.
    """
    assert len(centers) == len(sizes), "Number of centers and sizes must be the same"

    key = jax.random.PRNGKey(seed)
    keys = jax.random.split(key, len(centers))

    points = jnp.concatenate(
        [
            points_around(center, n_points=size, radius=radius, key=key)
            for center, size, key in zip(centers, sizes, keys)
        ],
        axis=0,
    )

    if shuffle:
        key, subkey = jax.random.split(key)
        points = jax.random.permutation(subkey, points)

    return points


def random_grid_points(
    n_points: int,
    dim: int,
    n_ticks: int,
    seed: int,
    spacing: int = 1,
):
    """
    Generate random grid points.
    """
    ticks = spacing * jnp.arange(n_ticks) - n_ticks / 2
    grid = jnp.stack(jnp.meshgrid(*[ticks] * dim), axis=-1).reshape(-1, dim)
    # TODO: do not generate all possible grid points, just sample

    key = jax.random.PRNGKey(seed)
    sample = jax.random.choice(key, grid, shape=(n_points,), replace=False)
    return sample


def partition(data: list, *, n_partitions: int):
    """
    Partition the data into `n_partitions` partitions.
    """

    n_points = len(data)
    partition_size = n_points // n_partitions + 1
    partitions = [
        data[i : i + partition_size] for i in range(0, n_points, partition_size)
    ]
    return partitions



import jax
import jax.numpy as jnp
from jax import lax


@jax.jit
def tanimoto(a: jnp.ndarray, b: jnp.ndarray) -> float:
    """
    Computes the Tanimoto distance between two vectors.
    """

    bitwise_and = jnp.sum(jnp.bitwise_and(a, b))
    bitwise_or = jnp.sum(jnp.bitwise_or(a, b))

    # Check for the case where both vectors are all zeros and return 0.0 in that case
    dist = jax.lax.cond(
        bitwise_or == 0,
        lambda _: 0.0,
        lambda _: 1 - (bitwise_and / bitwise_or),
        None,
    )

    return dist


@jax.jit
def euclidean(p1, p2):
    """
    Computes the Euclidean distance between two vectors.
    """
    return jnp.linalg.norm(p1 - p2)


def _is_in_bag(x: jnp.ndarray, X: jnp.ndarray) -> bool:
    """
    Checks if a point is already in the bag.
    """
    return (X == x).all(axis=1).any()


is_in_bag = jax.jit(_is_in_bag)


def _one_to_many_dists(x: jnp.ndarray, X: jnp.ndarray, dist: callable) -> jnp.ndarray:
    """
    Computes the distance between one point and many points.
    """
    dists_from_x_to_X = jax.vmap(dist, in_axes=(None, 0))(x, X)
    return dists_from_x_to_X


one_to_many_dists = jax.jit(_one_to_many_dists, static_argnames="dist")


def pairwise_distances(X, dist_fn):
    """Compute pairwise distances between points in X using a custom distance function."""
    N = X.shape[0]

    # Create a function to compute distances from one point to all other points
    def single_point_distances(x):
        return jax.vmap(lambda y: dist_fn(x, y))(X)

    # Use vmap to compute pairwise distances for all points
    distances = jax.vmap(single_point_distances)(X)

    return distances


def submatrix(X: jnp.ndarray, remove_row: int, remove_col: int) -> jnp.ndarray:
    """
    Returns a submatrix of `X` removing the specified row and column.
    """
    return jnp.delete(
        jnp.delete(X, remove_row, axis=0, assume_unique_indices=True),
        remove_col,
        axis=1,
        assume_unique_indices=True,
    )


def find_nasty_point(X: jnp.ndarray, dist_fn: callable, potential_fn: callable) -> int:
    """
    Find a point in `X` removing which would decrease the total potential the most.
    """

    dists = pairwise_distances(X, dist_fn)
    potentials = jax.vmap(potential_fn)(dists) / 2  # divide by 2 because of symmetry
    potentials = jnp.nan_to_num(
        potentials, posinf=0
    )  # replace diagonal elements with 0
    total_potential = potentials.sum()

    # Drop each point and compute the total potential without it
    def body_fun(i):
        return submatrix(potentials, remove_row=i, remove_col=i).sum()

    total_potentials_without_each_point = jax.vmap(body_fun)(jnp.arange(X.shape[0]))

    # Compute the decrease in the total potential
    deltas = total_potentials_without_each_point - total_potential

    # Find the point that would decrease the total potential the most (deltas are negative)
    nasty_point_idx = deltas.argmin()

    return nasty_point_idx


def _add_point_to_bag(
    x: jnp.ndarray,
    X: jnp.ndarray,
    dist: callable,
    n_neighbors: int,
    threshold: float = 0.0,
    approx_min: bool = True,
    power: float = -2,
) -> (jnp.ndarray, bool, int):
    """
    Adds one point to the bag, return the acceptance flag, the updated bag and the index of the replaced point (or -1
    if no point was replaced).
    """
    dists_from_x_to_X = _one_to_many_dists(x, X, dist)

    def below_threshold(X):
        return X, False, -1

    def above_threshold(X):
        # If there are less points in `X` than `n_neighbors`, use all points in `X`
        k_neigbour = min(n_neighbors, len(X))

        # Find closest points in `X` to `x`
        # TODO: test approx_min_k vs argpartition
        if approx_min:
            k_closest_points_dists, k_closest_points_indices = lax.approx_min_k(
                dists_from_x_to_X, k=k_neigbour
            )
        else:
            k_closest_points_dists, k_closest_points_indices = lax.top_k(
                -dists_from_x_to_X, k_neigbour
            )
            k_closest_points_dists = -k_closest_points_dists

        # Define a neighborhood of `x`
        N = jnp.concatenate((jnp.array([x]), X[k_closest_points_indices]))

        # Find a point in `N` removing which would decrease the total potential the most
        nasty_point_local_idx = find_nasty_point(N, dist, lambda d: jnp.power(d, power))

        # If the nasty point is not `x`, replace it with `x`
        is_accepted = nasty_point_local_idx > 0
        replace_idx = k_closest_points_indices[nasty_point_local_idx - 1]

        X = lax.cond(
            is_accepted,
            lambda X: X.at[replace_idx].set(x),
            lambda X: X,
            X,
        )

        return X, is_accepted, k_closest_points_indices[nasty_point_local_idx - 1]

    X, is_accepted, replace_idx = lax.cond(
        dists_from_x_to_X.min() > threshold,
        above_threshold,
        below_threshold,
        X,
    )

    return X, is_accepted, replace_idx


add_point_to_bag = jax.jit(
    _add_point_to_bag,
    static_argnames=["dist", "n_neighbors", "threshold", "approx_min"],
)


def _keep_last_changes(changes: jnp.ndarray) -> jnp.ndarray:
    """
    Given an array where each element represents whether a change occurred or
    not, -1 means no change, and a positive integer represents ID of changed
    object in some DB, return an array where only the last occurrence of each
    change is kept.

    Example:
    [ -1,  0,  2,  1,  1,  2,  2,  -1] <- identify unique changes
    [      0           1       2     ] <- only last changes are kept, -1s are ignored
    [ -1,  0, -1, -1,  1, -1,  2,  -1] <- all other elements are set to -1
    """
    # Take the last occurrence of each change
    changes_reversed = changes[::-1]
    unique_changes, unique_idxs = jnp.unique(
        changes_reversed, return_index=True, size=changes.shape[0]
    )

    def keep_only_unique_change(i, changes_reversed):
        return lax.cond(
            jnp.isin(i, unique_idxs),
            lambda _: changes_reversed,
            lambda _: changes_reversed.at[i].set(-1),
            changes_reversed,
        )

    changes_reversed = lax.fori_loop(
        0, changes_reversed.shape[0], keep_only_unique_change, changes_reversed
    )

    return changes_reversed[::-1]


keep_last_changes = jax.jit(_keep_last_changes)


def _add_points_to_bag(
    xs: jnp.ndarray,
    X: jnp.ndarray,
    dist: callable,
    n_neighbors: int,
    threshold: float,
) -> jnp.ndarray:
    # Initialize array to store the information about the changes
    changed_item_idxs = -jnp.ones(xs.shape[0], dtype=int)  # -1 means not altered

    def body_fun(i, args):
        X, altered_items_idxs = args
        X_new, _, altered_item_idx = _add_point_to_bag(
            xs[i], X, dist, n_neighbors=n_neighbors, threshold=threshold
        )

        # Record the index of the replaced point
        altered_items_idxs = altered_items_idxs.at[i].set(altered_item_idx)

        return X_new, altered_items_idxs

    X, changed_item_idxs = lax.fori_loop(
        0, xs.shape[0], body_fun, (X, changed_item_idxs)
    )
    # Some points might have been accepted and then replaced by another point
    changed_item_idxs = _keep_last_changes(changed_item_idxs)
    accepted_points_mask = changed_item_idxs >= 0

    return X, accepted_points_mask, changed_item_idxs


add_points_to_bag = _add_points_to_bag

add_points_to_bag = jax.jit(
    _add_points_to_bag,
    static_argnames=["dist", "n_neighbors", "threshold"],
)
